@@ -73,8 +73,8 @@ def get_completer(shell):
     try:
         return _SUPPORTED_COMPLETERS[shell]
     except KeyError:
-        raise NotImplementedError("shell (%s) must be in {%s}" %
-                                  (shell, ",".join(SUPPORTED_SHELLS)))
+        supported = ",".join(SUPPORTED_SHELLS)
+        raise NotImplementedError(f"shell ({shell}) must be in {supported}")
 
 
 @total_ordering
@@ -186,8 +186,8 @@ def get_bash_commands(root_parser, root_prefix, choice_functions=None):
 
             if hasattr(positional, "complete"):
                 # shtab `.complete = ...` functions
-                compgens.append("{}_pos_{}_COMPGEN={}".format(
-                    prefix, i, complete2pattern(positional.complete, "bash", choice_type2fn)))
+                comp_pattern = complete2pattern(positional.complete, "bash", choice_type2fn)
+                compgens.append(f"{prefix}_pos_{i}_COMPGEN={comp_pattern}")
 
             if positional.choices:
                 # choices (including subparsers & shtab `.complete` functions)
@@ -199,8 +199,7 @@ def get_bash_commands(root_parser, root_prefix, choice_functions=None):
                         # append special completion type to `compgens`
                         # NOTE: overrides `.complete` attribute
                         log.debug(f"Choice.{choice.type}:{prefix}:{positional.dest}")
-                        compgens.append("{}_pos_{}_COMPGEN={}".format(
-                            prefix, i, choice_type2fn[choice.type]))
+                        compgens.append(f"{prefix}_pos_{i}_COMPGEN={choice_type2fn[choice.type]}")
                     elif isinstance(positional.choices, dict):
                         # subparser, so append to list of subparsers & recurse
                         log.debug("subcommand:%s", choice)
@@ -216,7 +215,7 @@ def get_bash_commands(root_parser, root_prefix, choice_functions=None):
                                 new_nargs,
                             ) = recurse(
                                 positional.choices[choice],
-                                prefix + "_" + wordify(choice),
+                                f"{prefix}_{wordify(choice)}",
                             )
                             sub_subparsers.extend(new_subparsers)
                             sub_option_strings.extend(new_option_strings)
@@ -230,21 +229,21 @@ def get_bash_commands(root_parser, root_prefix, choice_functions=None):
                         this_positional_choices.append(str(choice))
 
                 if this_positional_choices:
-                    choices.append("{}_pos_{}_choices=('{}')".format(
-                        prefix, i, "' '".join(this_positional_choices)))
+                    choices_str = "' '".join(this_positional_choices)
+                    choices.append(f"{prefix}_pos_{i}_choices=('{choices_str}')")
 
             # skip default `nargs` values
             if positional.nargs not in (None, "1", "?"):
                 nargs.append(f"{prefix}_pos_{i}_nargs={positional.nargs}")
 
         if discovered_subparsers:
-            subparsers.append("{}_subparsers=('{}')".format(prefix,
-                                                            "' '".join(discovered_subparsers)))
+            subparsers_str = "' '".join(discovered_subparsers)
+            subparsers.append(f"{prefix}_subparsers=('{subparsers_str}')")
             log.debug(f"subcommands:{prefix}:{discovered_subparsers}")
 
         # optional arguments
-        option_strings.append("{}_option_strings=('{}')".format(
-            prefix, "' '".join(get_option_strings(parser))))
+        options_strings_str = "' '".join(get_option_strings(parser))
+        option_strings.append(f"{prefix}_option_strings=('{options_strings_str}')")
         for optional in parser._get_optional_actions():
             if optional == SUPPRESS:
                 continue
@@ -252,9 +251,9 @@ def get_bash_commands(root_parser, root_prefix, choice_functions=None):
             for option_string in optional.option_strings:
                 if hasattr(optional, "complete"):
                     # shtab `.complete = ...` functions
-                    compgens.append("{}_{}_COMPGEN={}".format(
-                        prefix, wordify(option_string),
-                        complete2pattern(optional.complete, "bash", choice_type2fn)))
+                    comp_pattern_str = complete2pattern(optional.complete, "bash", choice_type2fn)
+                    compgens.append(
+                        f"{prefix}_{wordify(option_string)}_COMPGEN={comp_pattern_str}")
 
                 if optional.choices:
                     # choices (including shtab `.complete` functions)
@@ -264,20 +263,21 @@ def get_bash_commands(root_parser, root_prefix, choice_functions=None):
                         # NOTE: overrides `.complete` attribute
                         if isinstance(choice, Choice):
                             log.debug(f"Choice.{choice.type}:{prefix}:{optional.dest}")
-                            compgens.append("{}_{}_COMPGEN={}".format(
-                                prefix, wordify(option_string), choice_type2fn[choice.type]))
+                            func_str = choice_type2fn[choice.type]
+                            compgens.append(
+                                f"{prefix}_{wordify(option_string)}_COMPGEN={func_str}")
                         else:
                             # simple choice
                             this_optional_choices.append(str(choice))
 
                     if this_optional_choices:
-                        choices.append("{}_{}_choices=('{}')".format(
-                            prefix, wordify(option_string), "' '".join(this_optional_choices)))
+                        this_choices_str = "' '".join(this_optional_choices)
+                        choices.append(
+                            f"{prefix}_{wordify(option_string)}_choices=('{this_choices_str}')")
 
                 # Check for nargs.
                 if optional.nargs is not None and optional.nargs != 1:
-                    nargs.append("{}_{}_nargs={}".format(prefix, wordify(option_string),
-                                                         optional.nargs))
+                    nargs.append(f"{prefix}_{wordify(option_string)}_nargs={optional.nargs}")
 
         # append recursion results
         subparsers.extend(sub_subparsers)
@@ -298,7 +298,7 @@ def complete_bash(parser, root_prefix=None, preamble="", choice_functions=None):
 
     See `complete` for arguments.
     """
-    root_prefix = wordify("_shtab_" + (root_prefix or parser.prog))
+    root_prefix = wordify(f"_shtab_{root_prefix or parser.prog}")
     subparsers, option_strings, compgens, choices, nargs = get_bash_commands(
         parser, root_prefix, choice_functions=choice_functions)
 
@@ -339,7 +339,7 @@ _shtab_replace_nonword() {
 # set default values (called for the initial parser & any subparsers)
 _set_parser_defaults() {
   local subparsers_var="${prefix}_subparsers[@]"
-  sub_parsers=${!subparsers_var}
+  sub_parsers=${!subparsers_var-}
 
   local current_option_strings_var="${prefix}_option_strings[@]"
   current_option_strings=${!current_option_strings_var}
@@ -356,13 +356,13 @@ _set_new_action() {
   current_action="${prefix}_$(_shtab_replace_nonword $1)"
 
   local current_action_compgen_var=${current_action}_COMPGEN
-  current_action_compgen="${!current_action_compgen_var}"
+  current_action_compgen="${!current_action_compgen_var-}"
 
   local current_action_choices_var="${current_action}_choices[@]"
-  current_action_choices="${!current_action_choices_var}"
+  current_action_choices="${!current_action_choices_var-}"
 
   local current_action_nargs_var="${current_action}_nargs"
-  if [ -n "${!current_action_nargs_var}" ]; then
+  if [ -n "${!current_action_nargs_var-}" ]; then
     current_action_nargs="${!current_action_nargs_var}"
   else
     current_action_nargs=1
@@ -384,8 +384,8 @@ ${root_prefix}() {
   local completing_word="${COMP_WORDS[COMP_CWORD]}"
   COMPREPLY=()
 
-  prefix=${root_prefix}
-  word_index=0
+  local prefix=${root_prefix}
+  local word_index=0
   _set_parser_defaults
   word_index=1
 
@@ -394,13 +394,13 @@ ${root_prefix}() {
   while [ $word_index -ne $COMP_CWORD ]; do
     local this_word="${COMP_WORDS[$word_index]}"
 
-    if [[ -n $sub_parsers && " ${sub_parsers[@]} " =~ " ${this_word} " ]]; then
+    if [[ -n $sub_parsers && " ${sub_parsers[@]} " == *" ${this_word} "* ]]; then
       # valid subcommand: add it to the prefix & reset the current action
       prefix="${prefix}_$(_shtab_replace_nonword $this_word)"
       _set_parser_defaults
     fi
 
-    if [[ " ${current_option_strings[@]} " =~ " ${this_word} " ]]; then
+    if [[ " ${current_option_strings[@]} " == *" ${this_word} "* ]]; then
       # a new action should be acquired (due to recognised option string or
       # no more input expected from current action);
       # the next positional action can fill in here
@@ -462,7 +462,7 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
     See `complete` for arguments.
     """
     prog = parser.prog
-    root_prefix = wordify("_shtab_" + (root_prefix or prog))
+    root_prefix = wordify(f"_shtab_{root_prefix or prog}")
 
     choice_type2fn = {k: v["zsh"] for k, v in CHOICE_FUNCTIONS.items()}
     if choice_functions:
@@ -474,14 +474,14 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
     def is_opt_multiline(opt):
         return isinstance(opt, OPTION_MULTI)
 
-    def format_optional(opt):
+    def format_optional(opt, parser):
+        get_help = parser._get_formatter()._expand_help
         return (('{nargs}{options}"[{help}]"' if isinstance(
             opt, FLAG_OPTION) else '{nargs}{options}"[{help}]:{dest}:{pattern}"').format(
                 nargs=('"(- : *)"' if is_opt_end(opt) else '"*"' if is_opt_multiline(opt) else ""),
-                options=("{{{}}}".format(",".join(opt.option_strings))
-                         if len(opt.option_strings) > 1 else '"{}"'.format("".join(
-                             opt.option_strings))),
-                help=escape_zsh(opt.help or ""),
+                options=("{{{}}}".format(",".join(opt.option_strings)) if len(opt.option_strings)
+                         > 1 else '"{}"'.format("".join(opt.option_strings))),
+                help=escape_zsh(get_help(opt) if opt.help else ""),
                 dest=opt.dest,
                 pattern=complete2pattern(opt.complete, "zsh", choice_type2fn) if hasattr(
                     opt, "complete") else
@@ -489,10 +489,11 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
                  "({})".format(" ".join(map(str, opt.choices)))) if opt.choices else "",
             ).replace('""', ""))
 
-    def format_positional(opt):
+    def format_positional(opt, parser):
+        get_help = parser._get_formatter()._expand_help
         return '"{nargs}:{help}:{pattern}"'.format(
             nargs={ONE_OR_MORE: "(*)", ZERO_OR_MORE: "(*):", REMAINDER: "(-)*"}.get(opt.nargs, ""),
-            help=escape_zsh((opt.help or opt.dest).strip().split("\n")[0]),
+            help=escape_zsh((get_help(opt) if opt.help else opt.dest).strip().split("\n")[0]),
             pattern=complete2pattern(opt.complete, "zsh", choice_type2fn) if hasattr(
                 opt, "complete") else
             (choice_type2fn[opt.choices[0].type] if isinstance(opt.choices[0], Choice) else
@@ -503,9 +504,9 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
     all_commands = {
         root_prefix: {
             "cmd": prog, "arguments": [
-                format_optional(opt)
+                format_optional(opt, parser)
                 for opt in parser._get_optional_actions() if opt.help != SUPPRESS] + [
-                    format_positional(opt) for opt in parser._get_positional_actions()
+                    format_positional(opt, parser) for opt in parser._get_positional_actions()
                     if opt.help != SUPPRESS and opt.choices is None],
             "help": (parser.description
                      or "").strip().split("\n")[0], "commands": [], "paths": []}}
@@ -518,7 +519,7 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
                 continue
             if not sub.choices or not isinstance(sub.choices, dict):
                 # positional argument
-                all_commands[prefix]["arguments"].append(format_positional(sub))
+                all_commands[prefix]["arguments"].append(format_positional(sub, parser))
             else:  # subparser
                 log.debug(f"choices:{prefix}:{sorted(sub.choices)}")
                 public_cmds = get_public_subcommands(sub)
@@ -530,18 +531,26 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
 
                     # optionals
                     arguments = [
-                        format_optional(opt) for opt in subparser._get_optional_actions()
+                        format_optional(opt, parser) for opt in subparser._get_optional_actions()
                         if opt.help != SUPPRESS]
 
                     # positionals
                     arguments.extend(
-                        format_positional(opt) for opt in subparser._get_positional_actions()
+                        format_positional(opt, parser)
+                        for opt in subparser._get_positional_actions()
                         if not isinstance(opt.choices, dict) if opt.help != SUPPRESS)
 
-                    new_pref = prefix + "_" + wordify(cmd)
+                    # help text
+                    formatter = subparser._get_formatter()
+                    backup_width = formatter._width
+                    formatter._width = 1234567 # large number to effectively disable wrapping
+                    desc = formatter._format_text(subparser.description or "").strip()
+                    formatter._width = backup_width
+
+                    new_pref = f"{prefix}_{wordify(cmd)}"
                     options = all_commands[new_pref] = {
-                        "cmd": cmd, "help": (subparser.description or "").strip().split("\n")[0],
-                        "arguments": arguments, "paths": [*paths, cmd]}
+                        "cmd": cmd, "help": desc.split("\n")[0], "arguments": arguments,
+                        "paths": [*paths, cmd]}
                     new_subcmds = recurse(subparser, new_pref, [*paths, cmd])
                     options["commands"] = {
                         all_commands[pref]["cmd"]: all_commands[pref]
@@ -564,7 +573,7 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
     def command_case(prefix, options):
         name = options["cmd"]
         commands = options["commands"]
-        case_fmt_on_no_sub = """{name}) _arguments -C ${prefix}_{name_wordify}_options ;;"""
+        case_fmt_on_no_sub = """{name}) _arguments -C -s ${prefix}_{name_wordify}_options ;;"""
         case_fmt_on_sub = """{name}) {prefix}_{name_wordify} ;;"""
 
         cases = []
@@ -575,14 +584,14 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
                            prefix=prefix))
         cases = "\n\t".expandtabs(8).join(cases)
 
-        return """\
+        return f"""\
 {prefix}() {{
   local context state line curcontext="$curcontext" one_or_more='(-)*' remainder='(*)'
 
   if ((${{{prefix}_options[(I)${{(q)one_or_more}}*]}} + ${{{prefix}_options[(I)${{(q)remainder}}*]}} == 0)); then  # noqa: E501
     {prefix}_options+=(': :{prefix}_commands' '*::: :->{name}')
   fi
-  _arguments -C ${prefix}_options
+  _arguments -C -s ${prefix}_options
 
   case $state in
     {name})
@@ -594,33 +603,34 @@ def complete_zsh(parser, root_prefix=None, preamble="", choice_functions=None):
       esac
   esac
 }}
-""".format(prefix=prefix, name=name, cases=cases)
+"""
 
     def command_option(prefix, options):
-        return """\
+        arguments = "\n  ".join(options["arguments"])
+        return f"""\
 {prefix}_options=(
   {arguments}
 )
-""".format(prefix=prefix, arguments="\n  ".join(options["arguments"]))
+"""
 
     def command_list(prefix, options):
         name = " ".join([prog, *options["paths"]])
-        commands = "\n    ".join('"{}:{}"'.format(escape_zsh(cmd), escape_zsh(opt["help"]))
+        commands = "\n    ".join(f'"{escape_zsh(cmd)}:{escape_zsh(opt["help"])}"'
                                  for cmd, opt in sorted(options["commands"].items()))
-        return """
+        return f"""
 {prefix}_commands() {{
   local _commands=(
     {commands}
   )
   _describe '{name} commands' _commands
-}}""".format(prefix=prefix, name=name, commands=commands)
+}}"""
 
-    preamble = """\
+    preamble = (f"""\
 # Custom Preamble
-{}
+{preamble.rstrip()}
 
 # End Custom Preamble
-""".format(preamble.rstrip()) if preamble else ""
+""" if preamble else "")
     # References:
     #   - https://github.com/zsh-users/zsh-completions
     #   - http://zsh.sourceforge.net/Doc/Release/Completion-System.html
@@ -655,7 +665,10 @@ def complete_tcsh(parser, root_prefix=None, preamble="", choice_functions=None):
     """
     Return tcsh syntax autocompletion script.
 
-    See `complete` for arguments.
+    root_prefix:
+      ignored (tcsh has no support for functions)
+
+    See `complete` for other arguments.
     """
     optionals_single = set()
     optionals_double = set()
@@ -669,34 +682,26 @@ def complete_tcsh(parser, root_prefix=None, preamble="", choice_functions=None):
     def get_specials(arg, arg_type, arg_sel):
         if arg.choices:
             choice_strs = ' '.join(map(str, arg.choices))
-            yield "'{}/{}/({})/'".format(
-                arg_type,
-                arg_sel,
-                choice_strs,
-            )
-        elif hasattr(arg, "complete"):
+            yield f"'{arg_type}/{arg_sel}/({choice_strs})/'"
+        elif hasattr(arg, 'complete'):
             complete_fn = complete2pattern(arg.complete, 'tcsh', choice_type2fn)
             if complete_fn:
-                yield "'{}/{}/{}/'".format(
-                    arg_type,
-                    arg_sel,
-                    complete_fn,
-                )
+                yield f"'{arg_type}/{arg_sel}/{complete_fn}/'"
 
     def recurse_parser(cparser, positional_idx, requirements=None):
-        log_prefix = '| ' * positional_idx
-        log.debug('%sParser @ %d', log_prefix, positional_idx)
+        log_prefix = "| " * positional_idx
+        log.debug("%sParser @ %d", log_prefix, positional_idx)
         if requirements:
-            log.debug('%s- Requires: %s', log_prefix, ' '.join(requirements))
+            log.debug("%s- Requires: %s", log_prefix, " ".join(requirements))
         else:
             requirements = []
 
         for optional in cparser._get_optional_actions():
-            log.debug('%s| Optional: %s', log_prefix, optional.dest)
+            log.debug("%s| Optional: %s", log_prefix, optional.dest)
             if optional.help != SUPPRESS:
                 # Mingle all optional arguments for all subparsers
                 for optional_str in optional.option_strings:
-                    log.debug('%s| | %s', log_prefix, optional_str)
+                    log.debug("%s| | %s", log_prefix, optional_str)
                     if optional_str.startswith('--'):
                         optionals_double.add(optional_str[2:])
                     elif optional_str.startswith('-'):
@@ -706,11 +711,11 @@ def complete_tcsh(parser, root_prefix=None, preamble="", choice_functions=None):
         for positional in cparser._get_positional_actions():
             if positional.help != SUPPRESS:
                 positional_idx += 1
-                log.debug('%s| Positional #%d: %s', log_prefix, positional_idx, positional.dest)
+                log.debug("%s| Positional #%d: %s", log_prefix, positional_idx, positional.dest)
                 index_choices[positional_idx][tuple(requirements)] = positional
                 if not requirements and isinstance(positional.choices, dict):
                     for subcmd, subparser in positional.choices.items():
-                        log.debug('%s| | SubParser: %s', log_prefix, subcmd)
+                        log.debug("%s| | SubParser: %s", log_prefix, subcmd)
                         recurse_parser(subparser, positional_idx, requirements + [subcmd])
 
     recurse_parser(parser, 0)
@@ -724,16 +729,21 @@ def complete_tcsh(parser, root_prefix=None, preamble="", choice_functions=None):
             # Multiple requirements
             nlist = []
             for nn, arg in ndict.items():
-                checks = [f'[ "$cmd[{iidx}]" == "{n}" ]' for iidx, n in enumerate(nn, start=2)]
                 if arg.choices:
-                    nlist.append('( {}echo "{}" || false )'.format(
-                        ' && '.join(checks + ['']),                 # Append the separator
-                        '\\n'.join(arg.choices),
-                    ))
-
+                    checks = [f'[ "$cmd[{iidx}]" == "{n}" ]' for iidx, n in enumerate(nn, start=2)]
+                    choices_str = "' '".join(arg.choices)
+                    checks_str = ' && '.join(checks + [f"echo '{choices_str}'"])
+                    nlist.append(f"( {checks_str} || false )")
             # Ugly hack
-            specials.append("'p@{}@`set cmd=($COMMAND_LINE); {}`@'".format(
-                str(idx), ' || '.join(nlist)))
+            nlist_str = ' || '.join(nlist)
+            specials.append(f"'p@{str(idx)}@`set cmd=($COMMAND_LINE); {nlist_str}`@'")
+
+    if optionals_double:
+        if optionals_single:
+            optionals_single.add('-')
+        else:
+            # Don't add a space after completing "--" from "-"
+            optionals_single = ('-', '-')
 
     return Template("""\
 # AUTOMATICALLY GENERATED by `shtab`
@@ -742,13 +752,13 @@ ${preamble}
 
 complete ${prog} \\
         'c/--/(${optionals_double_str})/' \\
-        'c/-/(${optionals_single_str} -)/' \\
+        'c/-/(${optionals_single_str})/' \\
         ${optionals_special_str} \\
         'p/*/()/'""").safe_substitute(
         preamble=("\n# Custom Preamble\n" + preamble +
                   "\n# End Custom Preamble\n" if preamble else ""), root_prefix=root_prefix,
-        prog=parser.prog, optionals_double_str=' '.join(optionals_double),
-        optionals_single_str=' '.join(optionals_single),
+        prog=parser.prog, optionals_double_str=' '.join(sorted(optionals_double)),
+        optionals_single_str=' '.join(sorted(optionals_single)),
         optionals_special_str=' \\\n        '.join(specials))
 
 
